# -*- coding: UTF-8 -*-
'''
This package contains classes related to Chess Games storage in SQLite files.
This package uses configuration file: SQLiteChessGC.txt installed in user home folder - see README.md file.
'''

# Python standard libraries
import os, os.path, sqlite3

# This package modules
from gamerec import ChessGame
from gamerec.storage import GameCollection
from gamerec.config import gamerec_cfgfolder

# Sqlite Encoding names: http://sqlite.com/pragma.html#pragma_encoding
# Python Encoding names: file:///usr/share/doc/python3.2/html/library/codecs.html#standard-encodings
_dictSqlite2Python = {'UTF-8':'utf_8',
                      'UTF-16':'utf_16',
                      'UTF-16le':'utf_16_le',
                      'UTF-16be':'utf_16_be'
                     }
_dictPython2Sqlite = {}
for (sqlcode, pycode) in _dictSqlite2Python.items():
    _dictPython2Sqlite[pycode] = sqlcode


class ChessGameCollection (GameCollection):
    def __init__ (self):
        self.data = []
        self.coding = "utf_8"
        configfile = os.path.join(gamerec_cfgfolder(), "SQLiteChessGC.txt")
        if os.access(configfile, os.R_OK):
            self._loadconfig(configfile)
        else:
            self.tags = ["Event","Site","Date","Round","Board","White","Black","WhiteElo","BlackElo","TimeControl","Result"]
    
    def _loadconfig (self, filepath):
        """ Load configuration file """
        self.tags = []
        file = open(filepath, "r")
        for line in file:
            line = line.strip()
            if line.startswith("%"):
                continue
            if not line:
                continue
            self.tags.append(line)
        file.close()
    
    def load (self, filepath, sort = ""):
        """ Load the data from SQLite database.
        - filepath - path to sqlite database file
        - sort - place games in collection in given order (comma separated expression list as in ORDER BY)
        """ 
        self.data = []
        con = sqlite3.connect(filepath)
        con.row_factory = sqlite3.Row
        cur = con.cursor()
        cur.execute('PRAGMA encoding')
        row = cur.fetchone()
        self.coding = _dictSqlite2Python[row[0]]
        sql = 'SELECT * FROM Games'
        if len(sort) > 0:
            sql += ' ORDER BY ' + sort
        cur.execute(sql)
        while True:
            row = cur.fetchone()
            if row == None:
                break
            game = ChessGame()
            for field in row.keys():
                if field == "game":
                    game.game = row[field]
                elif field == "xxx":
                    tags = row[field].split('|')
                    for tag in tags:
                        if not tag == "":
                            tagv = tag.split('=')
                            game.tags[tagv[0]] = tagv[1]
                elif not row[field] == None:
                    game.tags[field] = row[field]
            self.data.append(game)
        con.close()
    
    def games_iterator(self):
        """ Iterator thru games collection """
        for game in self.data:
            lgame = game
            vals = []
            for tag in self.tags:
                if tag in lgame.tags:
                    vals.append(lgame.tags[tag])
                    del lgame.tags[tag]
                else:
                    vals.append(None)
            xxx = ""
            for (tag, val) in lgame.tags.items():
                xxx += "{0}={1}|".format(tag, val)
            vals.append(xxx)
            vals.append(lgame.game)
            yield tuple(vals)
        
    def save (self, filepath, cleargames = False):
        """ Saves the data to SQLite database.
        - filepath - path to sqlite database file
        - cleargames - clear database before adding games
        """ 
        existingDb = os.access(filepath, os.R_OK + os.W_OK)
        con = sqlite3.connect(filepath)
        if not existingDb:
            if not self.coding == "utf_8":
                if self.coding in _dictPython2Sqlite:
                    coding = _dictPython2Sqlite[self.coding]
                else:
                    raise RuntimeError("SQLiteGameCollection.save: Unsupported coding \"{}\"".format(self.coding))
                con.execute('PRAGMA encoding="{0}"'.format(coding))
            sql = "CREATE TABLE IF NOT EXISTS Games("
            for tag in self.tags:
                sql += tag + ","
            sql += "xxx,game)"
            con.execute(sql)
            con.commit()
        elif cleargames:
            con.execute('DELETE FROM Games')
            con.commit()
        if len(self.data) > 0:
            sql = "INSERT INTO Games VALUES(" + "?," * (len(self.tags)+1) + "?)"
            #print("sql=",sql)
            con.executemany(sql, self.games_iterator())
            con.commit()
         


# -*- coding: utf-8 -*-
'''
This package contains classes related to games' collection storage.
'''

from collections import UserList

class GameCollection (UserList):
    """ Base class for concrete containers that are collection of 2 palyers' games """

    def load (self, source):
        ''' Load the data from source which can point to file or database locator or any other resource '''
        raise NotImplementedError

    def save (self, dest):
        ''' Saves the data to dest which can point to file or database locator or any other resource '''
        raise NotImplementedError


########################################################################
from gamerec.config import *
__author__  = gamerec_author
__status__  = gamerec_status
__version__ = gamerec_version
__date__    = gamerec_date



# -*- coding: utf-8 -*-
'''
This package contains generic classes for entire library.
'''

class Game:
    ''' Base class for games supported in this library. '''
    pass

class ChessGame (Game):
    ''' Class to support Chess Game. '''
    def __init__ (self):
        self.tags = dict()
        self.game = ""


########################################################################
from .config import *
__author__  = gamerec_author
__status__  = gamerec_status
__version__ = gamerec_version
__date__    = gamerec_date



"""This module contains package meta information and common functions."""

from os import name
from os.path import join, expanduser

# Set version number here
gamerec_author  = "Grzegorz Wierzchowski <gwierzchowski@wp.pl>"
gamerec_status  = '4 - Beta'
gamerec_version = '0.0.3'
gamerec_date    = '2016 January'

#Development Status :: 1 - Planning
#Development Status :: 2 - Pre-Alpha
#Development Status :: 3 - Alpha
#Development Status :: 4 - Beta
#Development Status :: 5 - Production/Stable
#Development Status :: 6 - Mature
#Development Status :: 7 - Inactive

def gamerec_cfgfolder():
    if name == 'posix':
        return join(expanduser("~"), ".config", "gamerec")
    if name == 'nt':
        return join(expanduser("~"), "AppData", "Local", "gamerec")
    return join(expanduser("~"), ".gamerec")



# -*- coding: utf-8 -*-
'''
This package contains classes related to games' collection storage in .pgn files.
'''


########################################################################
from gamerec.config import *
__author__  = gamerec_author
__status__  = gamerec_status
__version__ = gamerec_version
__date__    = gamerec_date



#!/usr/bin/env python3

# This is only for Python 2.7
from __future__ import print_function

import os, sys
import distutils.core, distutils.file_util

sys.path.insert(0, '.')
from gamerec_pkg.config import *


if os.name == 'posix':
    doclocation = 'share/doc/gamerec'
else:
    doclocation = 'doc/gamerec'

cfglocation = gamerec_cfgfolder()

with open('DESCR.md') as f:
    long_description = f.read()

dist = distutils.core.setup(name = 'gamerec',
    version = gamerec_version,
    description = "Library to process 2 players' games recordings (only chess now)",
    long_description = long_description,
    author = 'Grzegorz Wierzchowski',
    author_email = 'gwierzchowski@wp.pl',
    url = 'https://github.com/gwierzchowski/gamerec',
    packages = ['gamerec','gamerec.storage','gamerec.storage.pgnfile','gamerec.storage.sqlite'],
    package_dir = {'gamerec': 'gamerec_pkg'},
    data_files = [(doclocation, ['README.md', 'LICENSE.txt', 'DESCR.md', 'docs/Changelog.md']),
                  (cfglocation, ['conf/PGNChessGC.pgn', 'conf/SQLiteChessGC.txt'])],
    license = 'LGPLv3',
    platforms = 'any',
    classifiers = [
        'Development Status :: ' + gamerec_status,
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Games/Entertainment :: Board Games',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
    )

if sys.argv[1] in ['install']:
    # Unfotunately while installing using pip those print statements does not work on Windows
    print ('---------------------------------------------------------')
    isobj = dist.get_command_obj('install_data')
    readmedoc = None
    for doc in isobj.get_outputs():
        if 'README.md' in doc:
            readmedoc = doc
            break
    if readmedoc:
        print ('See %s file' % readmedoc)
    else:
        print ('See files deployed to documentation folder')
    print ('---------------------------------------------------------')



# -*- coding: UTF-8 -*-
'''
This package contains classes related to Chess Games storage in .pgn files.
This package uses configuration file: PGNChessGC.pgn installed in user home folder.

The standard of PGN file (point 4.1) states that a file is encoded using ISO 8859/1 (Latin 1) codes.
This class extends this a little and instroduces a special mark in the form of line
% vi: encoding=<encoding vi spec>
which should be at the very beggining of file.
Such line is by PGN standard a comment line that does not affect a games data.
But this module uses this as encoding specification for Python code while reading file.
This allows to have PGN file encoded in different standard and containg a national characters that are out of 
ISO 8859/1 coding shema. My experience is that SCID program properly inports and display utf-8 encoded
files with national characters, and I'm sorry but I want to have a game comments in my national langugage.
BTW: It is time to change PGN specification to use utf-8 instead of outdated ISO 8859/1.
In case of lack of this line ISO 8859/1 is assumed.

References:
Vi Encoding names: http://vimdoc.sourceforge.net/htmldoc/mbyte.html#encoding-names
PDN Specification1: http://www6.chessclub.com/help/PGN-spec
PDN Specification2: http://www.saremba.de/chessgml/standards/pgn/pgn-complete.htm
'''

# Python standard libraries
import re, os, os.path

# This package modules
from gamerec import ChessGame
from gamerec.storage import GameCollection
from gamerec.config import gamerec_cfgfolder

# Vi Encoding names: http://vimdoc.sourceforge.net/htmldoc/mbyte.html#encoding-names
# Python Encoding names: file:///usr/share/doc/python3.2/html/library/codecs.html#standard-encodings
_dictVi2Python = {'latin1':'latin_1',
                  'iso-8859-1':'latin_1',
                  'iso-8859-2':'iso8859_2',
                  'cp1250':'cp1250',
                  'utf-8':'utf_8',
                  'utf-16':'utf_16',
                  'utf-16 le':'utf_16_le',
                  'utf-16 be':'utf_16_be'
                 } #TODO: Finish this list.
_dictPython2Vi = {}
for (vicode, pycode) in _dictVi2Python.items():
    _dictPython2Vi[pycode] = vicode

# Vim modelines: http://vimdoc.sourceforge.net/htmldoc/options.html#modeline
_coding1re = re.compile(r"%\s+(?:vi:|vim:|ex:)\s*[\w\s:=]*(?:enc|encoding)=(\w[\w\-]*)[\w\s:=]*")

# Below RE for PGN file are kindly borrowed from http://www.pychess.org/ program (version 0.10.1).
# Many thanks to PyChess project !
_tagre = re.compile(r"\[([a-zA-Z]+)[ \t]+\"(.*?)\"\]")
_comre = re.compile(r"(?:\{.*?\})|(?:;.*?[\n\r])|(?:\$[0-9]+)", re.DOTALL)
_movre = re.compile(r"""
    (                   # group start
    (?:                 # non grouping parenthesis start
    [KQRBN]?            # piece
    [a-h]?[1-8]?        # unambiguous column or line
    x?                  # capture
    [a-h][1-8]          # destination square
    =?[QRBN]?           # promotion
    |O\-O(?:\-O)?       # castling
    |0\-0(?:\-0)?       # castling
    )                   # non grouping parenthesis end
    [+#]?               # check/mate
    )                   # group end
    [\?!]*              # traditional suffix annotations
    \s*                 # any whitespace
    """, re.VERBOSE)

# token categories
COMMENT_REST, COMMENT_BRACE, COMMENT_NAG, \
VARIATION_START, VARIATION_END, \
RESULT, FULL_MOVE, MOVE_COUNT, MOVE, MOVE_COMMENT = range(1,11)

_pattern = re.compile(r"""
    (\;.*?[\n\r])        # comment, rest of line style
    |(\{.*?\})           # comment, between {} 
    |(\$[0-9]+)          # comment, Numeric Annotation Glyph
    |(\()                # variation start
    |(\))                # variation end
    |(\*|1-0|0-1|1/2)    # result (spec requires 1/2-1/2 for draw, but we want to tolerate simple 1/2 too)
    |(
    ([0-9]{1,3}\s*[.]*\s*)?
    ([a-hxOoKQRBN1-8+#=]{2,7}
    |O\-O(?:\-O)?
    |0\-0(?:\-0)?)
    ([\?!]{1,2})*
    )    # move (full, count, move with ?!, ?!)
    """, re.VERBOSE | re.DOTALL)


class ChessGameCollection (GameCollection):
    '''Class to load and save Chess Game Collection in .pgn file.'''
    def __init__ (self):
        self.data = []
        self.coding = "latin_1"
        templatefile = os.path.join(gamerec_cfgfolder(), "PGNChessGC.pgn")
        if os.access(templatefile, os.R_OK):
            self._loadtemplate(templatefile)
        else:
            self.tags = ["Event","Site","Date","Round","White","Black","Result"]
            self.addtagspos = 6  #set to -1 for not placing additional tags
            self.gameformat = ""
    
    def _loadtemplate (self, filepath):
        """ Load file in PGN format and treat it as save template (order of tags)"""
        self.tags = []
        self.addtagspos = -1
        self.gameformat = ""
        file = open(filepath, "r")
        for line in file:
            line = line.lstrip()
            if line.startswith("%"):
                continue
            if not line:
                continue
            if line.startswith("["):
                tagmatch = _tagre.match(line)
                if tagmatch is not None:
                    if tagmatch.group(1) == "xxx":
                        self.addtagspos = len(self.tags)
                    else:
                        self.tags.append(tagmatch.group(1))
                continue
            self.gameformat += line
            break
        file.close()
    
    def load (self, filepath):
        """ Load the data from file in PGN format """
        self.data = []
        self.coding = "latin_1"
        codingKnown = False
        while True:
            file = open(filepath, "r", encoding=self.coding)
            codingChanged = False
            mode = 0 #0-begin/aftergame 1-tags 2-gametext
            game = ChessGame()
            for line in file:
                line = line.lstrip()
                if line.startswith("%"):
                    if not codingKnown:
                        codingmatch = _coding1re.match(line)
                        if codingmatch is not None:
                            if codingmatch.group(1) in _dictVi2Python:
                                self.coding = _dictVi2Python[codingmatch.group(1)]
                            else:
                                self.coding = codingmatch.group(1)
                            codingChanged = True
                            codingKnown = True
                            break
                    continue
                if mode != 2 and line.startswith("["):
                    codingKnown = True
                    mode = 1
                    tagmatch = _tagre.match(line)
                    if tagmatch is not None:
                        game.tags[tagmatch.group(1)] = tagmatch.group(2)
                    continue
                if not line:
                    if mode == 0:
                        pass
                    elif mode == 1:
                        pass
                    else:
                        self.data.append(game)
                        game = ChessGame()
                        mode = 0
                    continue
                game.game += line
                mode = 2
            if mode == 2:
                self.data.append(game)
            file.close()
            if not codingChanged:
                break
    
    def save (self, filepath, opmode = "a"):
        """ Saves the data to file in PGN format.
        Use opmode as file open flag (i.e. append to file or overwrite).
        """
        file = open(filepath, opmode, encoding=self.coding)
        if not self.coding == "latin_1":
            if self.coding in _dictPython2Vi:
                coding = _dictPython2Vi[self.coding]
            else:
                coding = self.coding
            print("% vi: encoding={0}\n".format(coding), file = file)
        for game in self.data:
            gametags = dict(game.tags)
            tags = []
            for tag in self.tags:
                if tag in gametags:
                    tags.append("[{0} \"{1}\"]".format(tag, gametags[tag]))
                    del gametags[tag]
                else:
                    tags.append("[{0} \"?\"]".format(tag))
            if self.addtagspos == -1:
                for tag in tags:
                    print(tag, file = file)
            else:
                for tag in tags[0:self.addtagspos]:
                    print(tag, file = file)
                for (tag, val) in gametags.items():
                    print("[{0} \"{1}\"]".format(tag, val), file = file)
                for tag in tags[self.addtagspos:]:
                    print(tag, file = file)
            print("", file = file)
            print(game.game, file = file)
        file.close()
         